package fb_config;
use vars qw(@ISA @EXPORT $FPATH $FLPATH);
@ISA = qw(Exporter);
@EXPORT = qw($FPATH $FLPATH);
$FPATH = "/usr/local/share/frozen-bubble";
$FLPATH = "/usr/local/libexec/frozen-bubble";
